﻿/*
 * XbeeNodeManager.c
 *
 * ノードの管理とパケットの管理はXBeeでの通信には必ずしも必要ないので、ファイルを分けました。
 * 1) 1パケットずつデータが欲しい場合や、
 * 2) RFデータとノードの情報がセットで欲しい場合や、
 * 3) 割り込みによるパケット処理2つ分よりも時間のかかる処理を実行する
 *   （時間のかかる処理を実行するとパケットの受信ロス（メモリが上書きされるので過去のメッセージを読み出せなくなる）が発生します）
 *   （当然、この処理はシリアルの受信割り込みより低位の処理であるものとします。高位だとハード的に受信しようがないため。）
 * 際に利用すると便利です。
 * 
 * 本コードを使用する際は、XbeeControl.cよりも先にインクルードして下さい。
 *
 * [memo] C++へ移植するならば、XbeeControlクラスを作って、それを継承した形にすると良いと思う。
 * Created: 2012/12/12 5:06:45
 * Author:  morishita
 * Histroy: 
 *          2013/1/20    バグを訂正した。
 *          2013/2/9     int XBee_GetRfPacketNum(void)を追加した。
 *          2013/2/23    Xbee_SetAddressAndPacketSize()でノードが正常に登録されないバグを修正
 *          2013/2/24    Xbee_GetOnePacket()にて、受信データが存在しない場合にsizeを0とすることを保証した。
 *          2013/3/5     Xbee_GetOnePacket()にて、格納可能なデータサイズを超えそうなケースがあったため引き渡しサイズを抑制した。
 */ 
#ifndef XBEE_NODE_MANAGER_C                         /* インクルードガード */
#define XBEE_NODE_MANAGER_C


//------header file include------------------------------------------------------
#include "fifo.h"
#include "XBee.h"
#include <stdbool.h>


//-----error check---------------------------------------------------------------
#ifdef XBEE_PROTOCOL
    #error "XbeeControl.c should be included after XbeeNodeManager.c ."
#endif


//------constant value-----------------------------------------------------------
#define XBEE_NODE_SIZE                      10              /* 記憶するノード数（アドレスを記憶するのに8 byte必要なので、そう多くは記憶できない） */
#define XBEE_PACKET_MEMORY_SIZE             10              /* 何個のパケットまでを覚えておくかを表す。取り得る範囲は1～XBee_BUFF_SIZE。 */
//------デバッグ用のマクロ-------------------デバッグ機能を利用しない場合はコメントアウトしてください。
//#define XBEE_MANAGER_DEBUG


//------構造体（本ソースコード内でのみ使用する分）-------------------------------
/************************************************************************************************/
/* パケットの送信元の番号                                                                       */
/************************************************************************************************/
typedef struct{
    int     packetSize;                                     // パケットのサイズ
    char    packetHost;                                     // パケットの送信元の番号（アドレスではなく、本ソース内で管理している番号を入れる。0xffは不明を表す。）
}XBEE_PACKET_INFO;

/**************************************************************************************************/
/* 未読のバッファ内に格納されているRFデータの送信元とデータサイズを覚えておく構造体==パケット情報 */
/**************************************************************************************************/
typedef struct{
    int                 rp;
    int                 wp;
    XBEE_PACKET_INFO    info[XBEE_PACKET_MEMORY_SIZE];
    bool                overflow;                           // true: オーバーフロー（未読のパケット数がXBEE_PACKET_MEMORY_SIZEを超えた場合）
    bool                error;                              // true: 受信データを把握できなくなっています。初期化して下さい。
}XBEE_PACKET_INFO_STORAGE;

/*************************************************************************************************/
/* 受信したパケットの送信元を記録する構造体                                                      */
/*************************************************************************************************/
typedef struct{
    int                 recordNum;                          // ノード記録数
    XBEE_NODE           node[XBEE_NODE_SIZE];               // ノード情報（アドレスなど）
    bool                overflow;                           // true: オーバーフロー（RFデータを送ってきたノードが多すぎて記憶しきれない場合）
}XBEE_NODE_STORAGE;

//------外部変数を宣言-----------------------------------------------------------
static XBEE_PACKET_INFO_STORAGE _xbeePacketStorageInfo;     // 受信パケット情報
static XBEE_NODE_STORAGE        _xbeeNode;                  // ノード情報

//------以下は、関数--------------------------------------

/************************************************************************/
/* 指定されたノード情報を初期化する                                     */
/************************************************************************/
static void Xbee_InitNodeUnit(XBEE_NODE *unit)
{
    unit->add.msb = 0ul;
    unit->add.lsb = 0ul;
    unit->pos.lat = 0l;
    unit->pos.lon = 0l;
    unit->Rssi = 0x00;
    unit->connection = false;
    return;
}

/************************************************************************/
/* マネージャ情報を初期化する                                           */
/* 具体的には、ノードとパケットの受信情報を初期化します。               */
/************************************************************************/
void Xbee_InitManager(void)
{
    int i;
    
    _xbeeNode.recordNum = 0;
    _xbeeNode.overflow = false;
    _xbeePacketStorageInfo.rp = 0;
    _xbeePacketStorageInfo.wp = 0;
    for(i = 0; i < XBEE_PACKET_MEMORY_SIZE; i++)
    {
        _xbeePacketStorageInfo.info[i].packetHost = 0xff;
        _xbeePacketStorageInfo.info[i].packetSize = 0;
    }
    _xbeePacketStorageInfo.overflow = false;
    _xbeePacketStorageInfo.error = false;
    for(i = 0; i < XBEE_NODE_SIZE; i++)
        (void)Xbee_InitNodeUnit(&_xbeeNode.node[i]);
    return;
}

/****************************************************************************************/
/* パケット送信元アドレスとパケットサイズを記録する                                     */
/*                                                                                      */
/* RFデータを受信した際にパケットの主とパケットのサイズを記憶させるのに使用します。     */
/* 引数は、速度優先のためポインタを渡す。                                               */
/* XbeeControl.cからの呼び出しを想定しています。                                        */
/****************************************************************************************/
void Xbee_SetAddressAndPacketSize(XBEE_ADD *add, int size)
{
    int i, t;
    char hostId = 0xff;
    //#ifdef XBEE_MANAGER_DEBUG
    //    char hoge[10];
    //#endif
    
    // ノード情報を走査し、ノードのIDを取り出す
    for (i = 0; i < _xbeeNode.recordNum; i++)
    {
        if (_xbeeNode.node[i].add.msb == add->msb && _xbeeNode.node[i].add.lsb == add->lsb)
        {
            hostId = (char)i;
            break;
        }
    }
    /*
    #ifdef XBEE_MANAGER_DEBUG
        XBee_send3("dbmg1: ");
        {
            (void)Unti_atoi(hoge, sizeof(hoge), (long)hostId);
            XBee_send3(hoge);
        }
        XBee_send3("\n");
    #endif
    */
    // 登録がなければ、登録を試みる
    if (hostId == 0xff && _xbeeNode.overflow == false)
    {
        if (_xbeeNode.recordNum < XBEE_NODE_SIZE)
        {
            _xbeeNode.node[i].add = *add;
            hostId = _xbeeNode.recordNum;
            _xbeeNode.recordNum++;
        }
        else
            _xbeeNode.overflow = true;                          // ノードの登録限界数を超えたことを知らせる
    }
    // ノード番号とパケットサイズを保存する
    t = (_xbeePacketStorageInfo.wp + 1) % XBEE_PACKET_MEMORY_SIZE;
    if (t != _xbeePacketStorageInfo.rp)                         // パケット登録オーバーフローを確認(true: non overflow)
    {
        //(void)rs2_puts("g\n"); // なぜかフリーズ
        _xbeePacketStorageInfo.overflow = false;
        _xbeePacketStorageInfo.info[_xbeePacketStorageInfo.wp].packetSize = size;
        _xbeePacketStorageInfo.info[_xbeePacketStorageInfo.wp].packetHost = hostId;
        _xbeePacketStorageInfo.wp = t;
        /*
        #ifdef XBEE_MANAGER_DEBUG
            //XBee_send3("dbmg2 size: ");
            //(void)Unti_atoi(hoge, sizeof(hoge), (long)size);
            //XBee_send3(hoge);
            //XBee_send3("\n");
            XBee_send3("dbmg2 t: ");
            (void)Unti_atoi(hoge, sizeof(hoge), (long)t);
            XBee_send3(hoge);
            XBee_send3("\n");
        #endif
        */
    }
    else
    {
        //#ifdef XBEE_MANAGER_DEBUG
        //    XBee_send3("dbmg3 over!\n ");
        //#endif
        // オーバーフローした際は、最後のパケットの情報を上書きする
        _xbeePacketStorageInfo.overflow = true;                 
        t = _xbeePacketStorageInfo.wp - 1;                      // 最後のインデックスを計算
        if(t < 0) 
            t += XBEE_PACKET_MEMORY_SIZE;
        if(_xbeePacketStorageInfo.info[t].packetSize < 30000)   // 変数のオーバーフローを防ぐ。もし溢れたら、データ構造そのものが不正となる。
        {
            //rs2_puts("hoge1\n");         // for debug
            _xbeePacketStorageInfo.info[t].packetSize += size;  // XBeeのパケットサイズをそのまま足してもintの最大値は超えないので、そのまま足す
        }
        else
        {
            //rs2_puts("hoge2\n");
            _xbeePacketStorageInfo.error = true;                // 初期化の必要性を通報する
        }        
        if(hostId != _xbeePacketStorageInfo.info[t].packetHost)
        {
            //rs2_puts("hoge3\n");
            _xbeePacketStorageInfo.info[t].packetHost = 0xff;   // 複数の送り主のパケットが結合する場合、送り主を不明とする
        }            
    }
    //(void)rs2_puts("g\n"); // なぜかフリーズ
    return;
}

/************************************************************************/
/* RFデータパケットの受信数を返す                                       */
/* もし、データの保存にエラーがあれば0を返します。                      */
/************************************************************************/
int XBee_GetRfPacketNum(void)
{
    (void)XBee_buff_check();                                    // 非割り込み処理を使用している場合に必要
    int t = _xbeePacketStorageInfo.wp - _xbeePacketStorageInfo.rp;
    if(t < 0)
        t += XBEE_PACKET_MEMORY_SIZE;
    if(_xbeePacketStorageInfo.error == true)
        t = 0;
    return t;
}

/*******************************************************************************************/
/* バッファから引き出されたデータ数の通報を受け、内部で保持しているパケット情報を更新する  */
/*                                                                                         */
/* XbeeControl.cからの呼び出しを想定しています。                                           */
/*******************************************************************************************/
void Xbee_GivePullOutNum(int size)
{
    while(size > 0 && XBee_GetRfPacketNum() > 0)
    {
        int readSize = _xbeePacketStorageInfo.info[_xbeePacketStorageInfo.rp].packetSize;   // 1パケット分のデータサイズを取得
        if (readSize > size)                                                                // 要求サイズがパケットサイズよりも小さい場合
        {
            _xbeePacketStorageInfo.info[_xbeePacketStorageInfo.rp].packetSize -= size;
            size = 0;
        }
        else                                                                                // パケットサイズを超えている場合は、
        {
            _xbeePacketStorageInfo.rp = (_xbeePacketStorageInfo.rp + 1) % XBEE_PACKET_MEMORY_SIZE;
            size -= readSize;
        }
    }
    // if(size > 0 && XBee_GetRfPacketNum() == 0) 通報されたデータ量以上のデータを引き出そうとしている状況なので、何かおかしい;
    return;
}

/************************************************************************/
/* 1パケット分のRFデータを返す                                          */
/* 受信データが存在しない場合：パラメータsize==0となります。            */
/************************************************************************/
XBEE_RF_PACKET Xbee_GetOnePacket(void)
{
    XBEE_RF_PACKET ans;
    int readSize, hostId;
    
    (void)XBee_buff_check();                            // 非割り込み処理を使用している場合に必要
    if (XBee_GetRfPacketNum() > 0)
    {
        readSize = _xbeePacketStorageInfo.info[_xbeePacketStorageInfo.rp].packetSize;
        hostId = _xbeePacketStorageInfo.info[_xbeePacketStorageInfo.rp].packetHost;
        ans.node = _xbeeNode.node[hostId];
        if(readSize > (int)sizeof(ans.dataPacket) - 1)  // 読み込みサイズの制限, -1はnull文字の分
            readSize = sizeof(ans.dataPacket) - 1;
        ans.size = readSize;
        (void)XBee_getsb(ans.dataPacket, readSize);
        ans.dataPacket[readSize] = 0x00;                // 文字列データの送受信時の簡単のために、ヌルで閉める
        _xbeePacketStorageInfo.overflow = false;
    }
    else
    {
        ans.size = 0;        
    }
    
    return ans;    
}

#endif
